#include <torch/extension.h>
#include <cuda.h>
#include <cuda_runtime.h>
#include <c10/cuda/CUDAGuard.h>

#define WARP_SIZE 32
#define TILE_SIZE 4  // Each warp processes 4 output elements simultaneously

__global__ void warp_tile_kernel(const float* __restrict__ x,
                                   const float* __restrict__ weight,
                                   const float* __restrict__ bias,
                                   float* __restrict__ out,
                                   int in_features,
                                   int out_features) {
    // Calculate indices
    int warp_id = threadIdx.x / WARP_SIZE;
    int lane_id = threadIdx.x % WARP_SIZE;
    int batch_idx = blockIdx.x;
    int base_out_idx = (blockIdx.y * blockDim.x / WARP_SIZE + warp_id) * TILE_SIZE;

    // Early exit if this warp's output indices are out of bounds
    if (base_out_idx >= out_features) return;

    // Each thread maintains TILE_SIZE partial sums
    float sums[TILE_SIZE] = {0.0f};
    
    // Input row pointer for current batch
    const float* x_row = x + batch_idx * in_features;
    
    // Process TILE_SIZE output features simultaneously
    #pragma unroll
    for (int tile = 0; tile < TILE_SIZE; tile++) {
        int out_idx = base_out_idx + tile;
        if (out_idx >= out_features) break;

        const float* w_row = weight + out_idx * in_features;

        // Use vectorized loads for better memory throughput
        int vec_count = in_features / 4;
        const float4* x_vec = reinterpret_cast<const float4*>(x_row);
        const float4* w_vec = reinterpret_cast<const float4*>(w_row);

        // Each thread processes its portion of the input vectors
        for (int i = lane_id; i < vec_count; i += WARP_SIZE) {
            float4 a = __ldg(x_vec + i);
            float4 b = __ldg(w_vec + i);
            sums[tile] += a.x * b.x + a.y * b.y + a.z * b.z + a.w * b.w;
        }

        // Handle remaining elements
        int rem_start = vec_count * 4;
        for (int i = rem_start + lane_id; i < in_features; i += WARP_SIZE) {
            sums[tile] += __ldg(x_row + i) * __ldg(w_row + i);
        }

        // Warp-level reduction using shuffle operations
        #pragma unroll
        for (int offset = WARP_SIZE/2; offset > 0; offset /= 2) {
            sums[tile] += __shfl_down_sync(0xffffffff, sums[tile], offset);
        }

        // First lane writes result
        if (lane_id == 0 && out_idx < out_features) {
            float result = sums[tile] + __ldg(bias + out_idx);
            out[batch_idx * out_features + out_idx] = (result > 0.0f) ? result : 0.0f;
        }
    }
}

torch::Tensor warp_tile_forward(torch::Tensor x,
                                  torch::Tensor weight,
                                  torch::Tensor bias) {
    TORCH_CHECK(x.is_cuda(), "x must be a CUDA tensor");
    TORCH_CHECK(weight.is_cuda(), "weight must be a CUDA tensor");
    TORCH_CHECK(bias.is_cuda(), "bias must be a CUDA tensor");

    int batch_size = x.size(0);
    int in_features = x.size(1);
    int out_features = weight.size(0);

    auto out = torch::empty({batch_size, out_features}, x.options());

    // Calculate grid and block dimensions
    int warps_per_block = 8;  // Tunable parameter
    int threads_per_block = warps_per_block * WARP_SIZE;
    int blocks_y = (out_features + (warps_per_block * TILE_SIZE) - 1) / (warps_per_block * TILE_SIZE);
    
    dim3 grid(batch_size, blocks_y);
    dim3 block(threads_per_block);

    cudaStream_t stream = c10::cuda::getCurrentCUDAStream();
    warp_tile_kernel<<<grid, block, 0, stream>>>(
        x.data_ptr<float>(),
        weight.data_ptr<float>(),
        bias.data_ptr<float>(),
        out.data_ptr<float>(),
        in_features,
        out_features
    );

    return out;
}

PYBIND11_MODULE(TORCH_EXTENSION_NAME, m) {
    m.def("forward", &warp_tile_forward, "GEMM with bias and ReLU using warp-level tiling and shuffle reduction (CUDA)");
}